package cz.neit.dm.config;


import cz.neit.dm.security.RestAuthenticationEntryPoint;
import cz.neit.dm.services.UserDetailsServiceImpl;
import cz.neit.security.acl.IPermissionsService;
import cz.neit.security.jwt.authentication.JwtAuthenticationFilter;
import cz.neit.security.jwt.authentication.JwtSpringConfiguration;
import cz.neit.security.permission.hints.simple.StaticHintEvaluator;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Import;
import org.springframework.http.HttpMethod;
import org.springframework.security.authentication.AuthenticationManager;
import org.springframework.security.config.BeanIds;
import org.springframework.security.config.annotation.authentication.builders.AuthenticationManagerBuilder;
import org.springframework.security.config.annotation.web.builders.HttpSecurity;
import org.springframework.security.config.annotation.web.configuration.EnableWebSecurity;
import org.springframework.security.config.annotation.web.configuration.WebSecurityConfigurerAdapter;
import org.springframework.security.config.http.SessionCreationPolicy;
import org.springframework.security.crypto.bcrypt.BCryptPasswordEncoder;
import org.springframework.security.web.authentication.www.BasicAuthenticationFilter;
import org.springframework.web.cors.CorsConfiguration;
import org.springframework.web.cors.CorsConfigurationSource;
import org.springframework.web.cors.UrlBasedCorsConfigurationSource;

import java.util.Arrays;

/**
 * Created by Jan Hrubý on 20.08.2018
 */
@Configuration
@EnableWebSecurity
@Import(JwtSpringConfiguration.class)
public class WebSecurityConfig extends WebSecurityConfigurerAdapter {

    private final UserDetailsServiceImpl userDetailsService;
    private final RestAuthenticationEntryPoint unauthorizedHandler;
    private final JwtAuthenticationFilter jwtAuthenticationFilter;

    @Autowired
    public WebSecurityConfig(UserDetailsServiceImpl userDetailsService, RestAuthenticationEntryPoint unauthorizedHandler, JwtAuthenticationFilter jwtAuthenticationFilter) {
        this.userDetailsService = userDetailsService;
        this.unauthorizedHandler = unauthorizedHandler;
        this.jwtAuthenticationFilter = jwtAuthenticationFilter;
    }

    @Bean
    @Autowired
    public StaticHintEvaluator staticHintEvaluator(IPermissionsService permissionsService) {
        return new StaticHintEvaluator(permissionsService);
    }


    @Override
    public void configure(AuthenticationManagerBuilder authenticationManagerBuilder) throws Exception {
        authenticationManagerBuilder
                .userDetailsService(userDetailsService)
                .passwordEncoder(passwordEncoder());
    }

    @Override
    protected void configure(HttpSecurity http) throws Exception {
        http
                .sessionManagement().sessionCreationPolicy(SessionCreationPolicy.STATELESS).and()
                .httpBasic().disable()
                .formLogin().disable()
                .logout().disable()
                .csrf().disable()
                .cors().and()
                .headers().frameOptions().sameOrigin().and();

// define exception handling
        http.exceptionHandling().authenticationEntryPoint(unauthorizedHandler);

        // Authorization configuration

        // Use for all request without login should be authenticated
        setForJWTTokenAuthentication(http);

        // Use old version
//        setForNoAuthentication(http);
    }


    @Bean(BeanIds.AUTHENTICATION_MANAGER)
    @Override
    public AuthenticationManager authenticationManagerBean() throws Exception {
        return super.authenticationManagerBean();
    }

    private void setForJWTTokenAuthentication(HttpSecurity http) throws Exception {
        // Allow login in
        http.authorizeRequests()
                .antMatchers("/login").permitAll()
                .antMatchers("/password/change/**").permitAll()
                .antMatchers("/reset-token/**").permitAll()
                .antMatchers("/files/download/**").permitAll()    //pusti me stahovat
                .antMatchers("/error/**").permitAll()
                .antMatchers("/static/**").permitAll()
                .antMatchers("/swagger-ui/**").permitAll()
                .antMatchers("/webjars/**").permitAll()
                .antMatchers("/v2/api-docs").permitAll()
                .antMatchers("/swagger-resources/**").permitAll()
                .antMatchers("/swagger-ui.html").permitAll();
        // Restrict other request only for authenticated users
        http.authorizeRequests().anyRequest().authenticated();
        // Register JwtAuthenticationFilter
        http.addFilterBefore(jwtAuthenticationFilter, BasicAuthenticationFilter.class);

    }

    private void setForNoAuthentication(HttpSecurity http) throws Exception {
        // Allow all request only for authenticated users
        http.authorizeRequests().anyRequest().permitAll();
    }

    @Bean
    CorsConfigurationSource corsConfigurationSource() {
        CorsConfiguration configuration = new CorsConfiguration();
        configuration.applyPermitDefaultValues();
        configuration.setAllowedMethods(Arrays.asList(
                HttpMethod.GET.name(),
                HttpMethod.HEAD.name(),
                HttpMethod.POST.name(),
                HttpMethod.PUT.name(),
                HttpMethod.DELETE.name(),
                HttpMethod.OPTIONS.name()
        ));
        configuration.addExposedHeader("Authorization");
        configuration.addExposedHeader("Accept-Language");

        UrlBasedCorsConfigurationSource source = new UrlBasedCorsConfigurationSource();
        source.registerCorsConfiguration("/**", configuration);
        return source;
    }


    @Bean
    public BCryptPasswordEncoder passwordEncoder() {
        return new BCryptPasswordEncoder(12);
    }


}
