package cz.neit.dm.controllers;

import com.fasterxml.jackson.databind.ObjectMapper;
import cz.neit.dm.dtos.mappers.FileMapper;
import cz.neit.dm.dtos.modeldto.FileDTO;
import cz.neit.dm.dtos.modeldto.DownloadTokenDTO;
import cz.neit.dm.entity.FileDM;
import cz.neit.dm.entity.Link;
import cz.neit.dm.exceptions.BadRequestException;
import cz.neit.dm.exceptions.ForbiddenException;
import cz.neit.dm.exceptions.InternalServerErrorException;
import cz.neit.dm.exceptions.ResourceNotFoundException;
import cz.neit.dm.helpers.EntityValidator;
import cz.neit.dm.io.FileManager;
import cz.neit.dm.security.SecurityContextUtils;
import cz.neit.dm.services.DownloadHistoryService;
import cz.neit.dm.services.FileService;
import cz.neit.dm.services.LinkService;
import cz.neit.dm.services.SideLinkService;
import cz.neit.security.permission.hints.annotation.AppendHints;
import cz.neit.security.permission.hints.annotation.PermissionHint;
import io.swagger.annotations.ApiOperation;
import lombok.extern.slf4j.Slf4j;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.util.StreamUtils;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.multipart.MultipartFile;

import javax.servlet.http.HttpServletResponse;
import javax.validation.Valid;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.file.Path;
import java.util.List;

@Slf4j
@RestController
@RequestMapping("files")
public class FileController {

    private final FileService fileService;
    private final DownloadHistoryService downloadHistoryService;
    private final FileManager fileManager;
    private final FileMapper fileMapper;
    private final LinkService linkService;
    private final SideLinkService sideLinkService;
    private final EntityValidator<FileDTO> fileDTOEntityValidator;

    public FileController(FileService fileService, DownloadHistoryService downloadHistoryService, FileManager fileManager, FileMapper fileMapper, LinkService linkService, SideLinkService sideLinkService, EntityValidator<FileDTO> fileDTOEntityValidator) {
        this.fileService = fileService;
        this.downloadHistoryService = downloadHistoryService;
        this.fileManager = fileManager;
        this.fileMapper = fileMapper;
        this.linkService = linkService;
        this.sideLinkService = sideLinkService;
        this.fileDTOEntityValidator = fileDTOEntityValidator;
    }

    @ApiOperation(value = "Get all files")
    @GetMapping("all")
    @PreAuthorize("hasPermission('fileDM', 'viewAll')")
    public List<FileDTO> showAllFiles() {
        return fileService.getAllFilesDTO();
    }

    @ApiOperation(value = "Get all files for current user (by groups and private files")
    @AppendHints
    @GetMapping("")
    @PreAuthorize("hasPermission('fileDM', 'showFiles')")
    public List<FileDTO> showAllFilesForCurrentUser() {
        return fileService.getAllFilesForCurrentUser();
    }

    @ApiOperation(value = "Get all files by AuthorId")
    @PermissionHint
    @GetMapping("author/{id}")
    @PreAuthorize("hasPermission('fileDM', 'changeAuthor')")
    public List<FileDTO> getAllFilesDTOByAuthorId(@PathVariable Long id) {
        return fileService.getFilesDTOByAuthorId(id);
    }

    @ApiOperation(value = "Get file by ID")
    @PermissionHint
    @GetMapping("{id}")
    @PreAuthorize("hasPermission(#id, 'fileDM', 'viewDetail')")
    public FileDTO getFileDTO(@PathVariable Long id) {
        return fileService.getFileDetailsDTO(id);
    }

    @ApiOperation(value = "Get file by ID, only a few params about File, which can be updated")
    @PermissionHint
    @GetMapping("{id}/update")
    @PreAuthorize("hasPermission(#id, 'fileDM', 'update')")
    public FileDTO getFileForUpdate(@PathVariable Long id) {
        return fileMapper.fileDmToFileDTO(fileService.getFileDetails(id));
    }

    @ApiOperation(value = "Get file by ID with substringed link description ")
    @AppendHints
    @GetMapping("{id}/detail")
    @PreAuthorize("hasPermission(#id, 'fileDM', 'viewDetail')")
    public FileDTO getFileDetailWithDesc(@PathVariable Long id) {
        return fileService.getFileDetailsWithDescription(id);
    }

    @ApiOperation(value = "Update file")
    @PermissionHint
    @PutMapping("{id}")
    @PreAuthorize("hasPermission(#id, 'fileDM', 'update')")
    public FileDTO updateFile(@Valid @RequestBody FileDTO fileDTO, @PathVariable Long id) {
        return fileService.updateFile(fileDTO, id);
    }

    @ApiOperation(value = "Get boolean if file is shared")
    @GetMapping("{id}/shared")
    @PreAuthorize("hasPermission(#id, 'fileDM', 'isShared')")
    public boolean isFileShared(@PathVariable Long id) {
        return fileService.isShared(id);
    }

    @ApiOperation(value = "Get confirmation about date conflict between deleteDate and link.validUntil")
    @PermissionHint
    @PostMapping("{id}/validexpiration")
    @PreAuthorize("hasPermission(#id, 'fileDM', 'update')")
    public boolean isFileDeleteDateAfterLinkValidUntil(@Valid @RequestBody FileDTO fileDTO, @PathVariable Long id) {
        return fileService.isAnyFileLinkIncorrect(fileMapper.fileDtoToFileDM(fileDTO));
    }

    @ApiOperation(value = "Get boolean if user (userId) has upload some files under the group (groupId")
    @GetMapping("{userId}/has-files/{groupId}")
    public boolean isAuthorHasFilesInGroup(@PathVariable Long userId, @PathVariable Long groupId) {
        return fileService.isAuthorHasFilesInGroup(userId, groupId);
    }

    /**
     * Deleting File from FileSystem and Database
     *
     * @param id
     * @throws InternalServerErrorException & BadRequestException
     */
    @ApiOperation(value = "Delete file by ID")
    @PermissionHint
    @DeleteMapping("{id}")
    @PreAuthorize("hasPermission(#id, 'fileDM', 'delete')")
    public void deleteFile(@PathVariable Long id) {
        if (!fileService.isShared(id)) {
            log.debug("Deleting file by id: " + id + " from filesystem.");
            Path path = fileManager.getFilePath(fileService.getFileDetails(id));
            fileService.deleteFile(id, path);
        } else {
            throw new BadRequestException("Soubor nemůže být smazán.");
        }
    }


    @ApiOperation(value = "Get boolean is file can be uploaded to app, no parameter is missing")
    @PostMapping("validate")
    @PreAuthorize("hasPermission('fileDM', 'create')")
    public boolean validateForSingleFileUpload(@RequestParam("file") MultipartFile file,
                                               @RequestParam("dto") String serializedFileDTO) {
        log.debug("Validation for upload file STARTED");
        try {

            FileDTO fileDTO = new ObjectMapper().readValue(serializedFileDTO, FileDTO.class);

            fileDTOEntityValidator.validate(fileDTO)
                    .forEach(constraintViolation -> fileDTOEntityValidator
                            .sendErrorMessageAndThrowBadRequest(constraintViolation.getMessage())
                    );
            log.debug("Validation is OVER and everything is okey.");
            return true;
        } catch (IOException e) {
            log.error(String.valueOf(e));
            throw new InternalServerErrorException("Vyskytl se neočekávaný problém s nahráním souboru na server. Kontaktujte svého administrátora.", e);
        }
    }

    /**
     * Deserialize fileDTO from Json, then create file -> copy stream and close both streams
     *
     * @param file
     * @param serializedFileDTO
     * @return return result of the upload (FileDTO)
     * @throws InternalServerErrorException
     */
    @ApiOperation(value = "Post a new file to app and save MultipartFile to FileSystem")
    @PostMapping("")
    @PreAuthorize("hasPermission('fileDM', 'create')")
    public FileDTO singleFileUpload(@RequestParam("file") MultipartFile file,
                                    @RequestParam("dto") String serializedFileDTO) {
        try {
            FileDTO fileDTO = new ObjectMapper().readValue(serializedFileDTO, FileDTO.class);

            fileDTOEntityValidator.validate(fileDTO)
                    .forEach(constraintViolation -> fileDTOEntityValidator
                            .sendErrorMessageAndThrowBadRequest(constraintViolation.getMessage())
                    );

            FileDM fileDM = fileService.createFileWithoutLink(fileDTO, file);
            log.debug("Saving file " + fileDM.getFilename() + " by id: " + fileDM.getId() + " to filesystem.");
            try (InputStream is = file.getInputStream();
                 OutputStream os = fileManager.getFileOutputStream(fileDM)) {
                StreamUtils.copy(is, os);
            }
            return fileMapper.fileDmToFileDTO(fileDM);
        } catch (IOException e) {
            log.error(String.valueOf(e));
            throw new InternalServerErrorException("Vyskytl se neočekávaný problém s nahráním souboru na server. Kontaktujte svého administrátora.", e);
        }
    }

    /**
     * For downloading file from APP
     *
     * @param id
     * @return token for downloading back to APP
     * @throws ForbiddenException
     */
    @ApiOperation(value = "Get DownloadTokenDTO for downloading file from APP")
    @PermissionHint
    @GetMapping("{id}/token")
    @PreAuthorize("hasPermission(#id, 'fileDM', 'download')")
    public DownloadTokenDTO getToken(@PathVariable Long id) {
        if (fileService.isDownloadable(id)) {
            downloadHistoryService.addNewDownloadHistory(SecurityContextUtils.getCurrentLoggedUser(), fileService.getFileDetails(id));
            return fileService.getDownloadTokenForFile(id);
        } else {
            throw new ForbiddenException("Chybí vám dostatečná práva ke stažení.");
        }
    }


    /**
     * @param token
     * @return
     * @throws InternalServerErrorException
     */
    @ApiOperation(value = "Download file")
    @GetMapping("download/token/{token}")
    public void downloadFileUnsecured(@PathVariable String token,
                                      HttpServletResponse httpServletResponse) throws IOException {

        FileDM fileDM = fileService.getFileByTokenFromAPP(token);
        this.downloadByOutputStreamHelper(httpServletResponse, fileDM);
    }

    /**
     * Download file from API endpoint via SIDELINK
     *
     * @param sideLinkToken
     * @param httpServletResponse
     * @return
     * @throws InternalServerErrorException
     */
    @ApiOperation(value = "Download file from link")
    @GetMapping("download/link/{sideLinkToken}")
    public void downloadFileFromLink(@PathVariable String sideLinkToken,
                                     HttpServletResponse httpServletResponse) throws IOException {

        Link link = sideLinkService.getSideLinkByToken(sideLinkToken).getLink();
        FileDM fileDM = link.getFile();

        try {
            if (!linkService.isDownloadsCountLimitExceeded(link) && linkService.isValidUntilDateExceeded(link)) {
                log.debug("Downloading file from URL by id: " + fileDM.getId());
                fileService.setNewLastDownloadViaLimit(fileDM.getId());

                this.downloadByOutputStreamHelper(httpServletResponse, fileDM);

                downloadHistoryService.addNewDownloadHistory(sideLinkToken);
                linkService.setPlusActualDownloadsCount(link);


            } else if (linkService.isDownloadsCountLimitExceeded(link)) {
                try {
                    httpServletResponse.sendRedirect("/error/dl-exceeded");
                } catch (IOException e) {
                    throw new InternalServerErrorException("IOException", e);
                }
            } else if (!linkService.isValidUntilDateExceeded(link)) {
                try {
                    httpServletResponse.sendRedirect("/error/time-exceeded");
                } catch (IOException e) {
                    throw new InternalServerErrorException("IOException", e);
                }
            } else {
                throw new InternalServerErrorException("Nastala neočekávaná vyjímka. Kontaktujte prosím svého administrátora.");
            }
        } catch (ResourceNotFoundException ex) {
            log.debug("Link or SideLink no exists", ex);
            try {
                httpServletResponse.sendRedirect("/error/link-invalidate");
            } catch (IOException e) {
                throw new InternalServerErrorException("IOException", e);
            }
        } catch (Exception e) {
            log.error("An error occurred while downloading");
            httpServletResponse.sendRedirect("/error/download-error");
        }

        if (fileService.isAllLimitsExceeded(fileDM.getId(), link)) {
            Path path = fileManager.getFilePath(fileDM);
            fileService.deleteFile(fileDM.getId(), path);
        }
    }


    private void downloadByOutputStreamHelper(HttpServletResponse httpServletResponse, FileDM fileDM) throws IOException {
        httpServletResponse.setContentType("application/force-download");
        httpServletResponse.setHeader("Content-Disposition", "attachment; filename=" + fileDM.getFilename());
        fileManager.downloadFile(fileDM.getId(), httpServletResponse.getOutputStream());
    }
}
