package cz.neit.dm.controllers;

import cz.neit.dm.dtos.modeldto.GroupDTO;
import cz.neit.dm.dtos.modeldto.GroupSimpleDTO;
import cz.neit.dm.dtos.modeldto.UserSimpleDTO;
import cz.neit.dm.entity.GroupMembership;
import cz.neit.dm.services.GroupMembershipService;
import cz.neit.dm.services.GroupService;
import cz.neit.dm.services.UserService;
import cz.neit.security.permission.hints.annotation.AppendHints;
import cz.neit.security.permission.hints.annotation.PermissionHint;
import lombok.extern.slf4j.Slf4j;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.*;

import javax.validation.Valid;
import java.util.List;

@Slf4j
@RestController
@RequestMapping("groups")
public class GroupController {
    private final GroupService groupService;
    private final GroupMembershipService groupMembershipService;
    private final UserService userService;

    public GroupController(GroupService groupService, GroupMembershipService groupMembershipService, UserService userService) {
        this.groupService = groupService;
        this.groupMembershipService = groupMembershipService;
        this.userService = userService;
    }

    @AppendHints
    @GetMapping("")
    @PreAuthorize("hasPermission('group', 'showGroups')")
    public List<GroupDTO> getListOfGroupsCurrentUser() {
        return groupService.getAllGroupsByCurrentUser();
    }

    @GetMapping("unassigned/{userId}")
    @PreAuthorize("hasPermission('group', 'showUnassigned')")
    public List<GroupDTO> getListOfUnassignedGroups(@PathVariable Long userId){
        return groupService.getAllUnassignedGroups(userId);
    }

    @GetMapping("{groupId}/managers/{userId}")
    @PreAuthorize("hasPermission(#groupId, 'group', 'showManagers')")
    public List<UserSimpleDTO> getListOfGroupManagers(@PathVariable Long groupId, @PathVariable Long userId) {
        return groupMembershipService.getManagersOfGroupForNewAuthor(groupId, userId);
    }

    @GetMapping("managers")
    @PreAuthorize("hasPermission('group', 'changeAuthor')")
    public List<UserSimpleDTO> getListOfManagers() {
        return groupMembershipService.getAllManagers();
    }

    @AppendHints
    @GetMapping("{id}")
    @PreAuthorize("hasPermission(#id, 'group', 'viewDetail')")
    public GroupDTO getGroupById(@PathVariable Long id){
        return groupService.getGroupDetailsDTO(id);
    }

    @AppendHints
    @GetMapping("/name/{id}")
    @PreAuthorize("hasPermission('group', 'showGroups')")
    public GroupDTO getGroupByName(@PathVariable String id){
        return groupService.getGroupDetailsByNameDTO(id);
    }


    //----Assign and UnAssign users and managers from group
    @PostMapping("{groupId}/users/{userId}")
    @PreAuthorize("hasPermission(#groupId,'group', 'assignUser')")
    public GroupMembership assignUserToGroup(@PathVariable Long groupId, @PathVariable Long userId) {
        return groupMembershipService.assignUserToGroup(groupId,userId);
    }

    @PostMapping("{groupId}/managers/{userId}")
    @PreAuthorize("hasPermission(#groupId,'group', 'assignManager')")
    public GroupMembership assignManagerToGroup(@PathVariable Long groupId, @PathVariable Long userId) {
        return groupMembershipService.assignManagerToGroup(groupId, userId);
    }

    @GetMapping("{groupId}/users/{userId}/has-files")
    @PreAuthorize("hasPermission(#groupId, 'group', 'unassignUser')")
    public boolean hasUserFilesInGroup(@PathVariable Long groupId, @PathVariable Long userId) {
        return groupMembershipService.hasUserFilesInGroup(groupId,userId);
    }


    @DeleteMapping("{groupId}/users/{userId}")
    @PreAuthorize("hasPermission(#groupId, 'group', 'unassignUser')")
    public void deleteUserFromGroupSimple(@PathVariable Long groupId, @PathVariable Long userId) {
        groupMembershipService.deleteGroupMembership(groupId, userId,null);
    }

    @DeleteMapping("{groupId}/users/{userId}/new-author/{newAuthorId}")
    @PreAuthorize("hasPermission(#groupId, 'group', 'unassignUser')")
    public void deleteUserFromGroup(@PathVariable Long groupId, @PathVariable Long userId, @PathVariable Long newAuthorId) {
        groupMembershipService.deleteGroupMembership(groupId, userId, userService.getUserDetail(newAuthorId));
    }

    @DeleteMapping("{groupId}/managers/{userId}/new-author/{newAuthorId}")
    @PreAuthorize("hasPermission(#groupId, 'group', 'unassignManagerAndKick')")
    public void deleteManagerFromGroup(@PathVariable Long groupId, @PathVariable Long userId, @PathVariable Long newAuthorId) {
        groupMembershipService.deleteGroupMembership(groupId, userId, userService.getUserDetail(newAuthorId));
    }

    @DeleteMapping("{groupId}/managers/{userId}/move")
    @PreAuthorize("hasPermission(#groupId, 'group', 'unassignManager')")
    public void unassignManagerFromGroup(@PathVariable Long groupId, @PathVariable Long userId) {
        groupMembershipService.deleteManagerGroupMembership(groupId, userId);
        groupMembershipService.assignUserToGroup(groupId, userId);
    }




    //------------------------------------------------------------------------------------------------------------------

    @PostMapping("")
    @PreAuthorize("hasPermission('group', 'create')")
    public GroupSimpleDTO createGroup(@Valid @RequestBody GroupSimpleDTO groupSimpleDTO){
        return groupService.createGroup(groupSimpleDTO);
    }

    @PermissionHint
    @PutMapping("{id}")
    @PreAuthorize("hasPermission(#id, 'group', 'update')")
    public GroupSimpleDTO updateGroup(@Valid @RequestBody GroupSimpleDTO groupSimpleDTO, @PathVariable Long id) {
        return groupService.updateGroup(id, groupSimpleDTO);
    }

    @PermissionHint
    @DeleteMapping("{id}")
    @PreAuthorize("hasPermission(#id, 'group', 'delete')")
    public void deleteGroup(@PathVariable Long id) {
        groupService.deleteGroup(id);
    }

    @GetMapping("{id}/can-delete")
    @PreAuthorize("hasPermission(#id, 'group', 'delete')")
    public boolean isDeletable(@PathVariable Long id) {
        return groupService.isDeletable(id);
    }
}
