package cz.neit.dm.controllers;

import cz.neit.dm.dtos.mappers.LinkMapper;
import cz.neit.dm.dtos.modeldto.LinkDTO;
import cz.neit.dm.dtos.modeldto.LinkValidationDTO;
import cz.neit.dm.entity.Link;
import cz.neit.dm.mail.EmailService;
import cz.neit.dm.services.LinkService;
import cz.neit.security.permission.hints.annotation.AppendHints;
import cz.neit.security.permission.hints.annotation.PermissionHint;
import lombok.extern.slf4j.Slf4j;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.*;

import javax.validation.Valid;
import java.util.List;

/**
 * Created by jhruby on 05.09.2018
 */
@Slf4j
@RestController
@RequestMapping("links")
public class LinkController {

    private final LinkService linkService;
    private final EmailService emailService;
    private final LinkMapper linkMapper;

    public LinkController(LinkService linkService, EmailService emailService, LinkMapper linkMapper) {
        this.linkService = linkService;
        this.emailService = emailService;
        this.linkMapper = linkMapper;
    }

    @AppendHints
    @GetMapping("")
    @PreAuthorize("hasPermission('link', 'showLinks')")
    public List<LinkDTO> getLinksForCurrentUser(){
        return linkService.getLinksForCurrentUser();
    }

    @AppendHints
    @GetMapping("{id}")
    @PreAuthorize("hasPermission(#id, 'link', 'viewDetail')")
    public LinkDTO getLink(@PathVariable Long id){
        return linkService.getLinkDetailDtoWithRecipients(id);
    }

    @PermissionHint
    @GetMapping("{id}/update")
    @PreAuthorize("hasPermission(#id, 'link', 'update')")
    public LinkDTO getLinkForUpdate(@PathVariable Long id){
        return linkService.getLinkDetailDtoWithRecipients(id);
    }

    @PostMapping("")
    @PreAuthorize("hasPermission('link', 'create')")
    public LinkDTO addLink(@Valid @RequestBody LinkDTO linkDTO) {
        return linkService.createLink(linkDTO);
    }

    @PermissionHint
    @PostMapping("validexpiration")
    @PreAuthorize("hasPermission('link', 'create')")
    public LinkValidationDTO isAnyLinkFileIncorrectCreate(@Valid @RequestBody LinkDTO linkDTO) {
        return linkService.isAnyLinkFileIncorrect(linkMapper.linkDtoToLink(linkDTO));
    }

    @PermissionHint
    @PostMapping("{id}/validexpiration")
    @PreAuthorize("hasPermission(#id, 'link', 'update')")
    public LinkValidationDTO isAnyLinkFileIncorrectUpdate(@Valid @RequestBody LinkDTO linkDTO, @PathVariable Long id) {
        return linkService.isAnyLinkFileIncorrect(linkMapper.linkDtoToLink(linkDTO));
    }

    @PermissionHint
    @PutMapping("{id}")
    @PreAuthorize("hasPermission(#id, 'link', 'update')")
    public LinkDTO updateLink(@Valid @RequestBody LinkDTO linkDTO, @PathVariable Long id) {
        return linkService.updateLink(id,linkDTO);
    }

    @AppendHints
    @PermissionHint
    @PutMapping("{id}/share")
    @PreAuthorize("hasPermission(#id, 'link', 'shareLink')")
    public LinkDTO shareLink(@PathVariable Long id)  {
        return linkService.shareLink(id);
    }

    @PermissionHint
    @PutMapping("{id}/resend")
    @PreAuthorize("hasPermission(#id, 'link', 'resendEmail')")
    public void resendEmail(@PathVariable Long id) {
        Link link = linkService.getLinkDetails(id);
        emailService.sendMimeMessage(link, linkService.getRecipientsByLinkId(link.getId()));
    }

    @AppendHints
    @PermissionHint
    @PutMapping("{id}/disable")
    @PreAuthorize("hasPermission(#id, 'link', 'disableLink')")
    public LinkDTO disableLink(@PathVariable Long id) {
        return linkService.disableLink(id);
    }

    @PermissionHint
    @DeleteMapping("{id}")
    @PreAuthorize("hasPermission(#id, 'link', 'delete')")
    public void deleteLink(@PathVariable Long id) {
        linkService.deleteLink(id);
    }






}
