package cz.neit.dm.io;

import cz.neit.dm.dtos.mappers.FileMapper;
import cz.neit.dm.entity.FileDM;
import cz.neit.dm.exceptions.InternalServerErrorException;
import cz.neit.dm.services.FileService;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;

import java.io.*;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.text.Normalizer;

/**
 * Created by jhruby on 29.08.2018
 */
@Slf4j
@Component
public class FileManager {

    private final FileService fileService;
    private final FileMapper fileMapper;

    @Value("${app.files.storageDir}")
    private String BASE_URL;


    public FileManager(FileService fileService, FileMapper fileMapper) {
        this.fileService = fileService;
        this.fileMapper = fileMapper;
    }

    public FileOutputStream getFileOutputStream(FileDM fileDM) throws InternalServerErrorException {
        try {
            createDirectoryIfNotExists(getDirectoryPath(fileDM));
            return new FileOutputStream(getFilePath(fileDM).toString());
        } catch (IOException e) {
            log.error(String.valueOf(e));
            throw new InternalServerErrorException("Soubor nebyl správně umístěn. Obraťte se na svého administrátora.", e);
        }
    }

    private FileInputStream getFileInputStream(FileDM fileDM) throws InternalServerErrorException {
        try {
            Path filePath = getFilePath(fileDM);
            return new FileInputStream(filePath.toString());
        } catch (FileNotFoundException e) {
            log.error(String.valueOf(e));
            throw new InternalServerErrorException("Soubor nebyl nalezen. Obraťte se na svého administrátora.", e);
        }
    }

    private String getFileNameForPath(FileDM fileDM) {
        String fileName = (fileDM.getId() + fileDM.getFilename()).toLowerCase().replaceAll("\\s", "");
        fileName = Normalizer.normalize(fileName, Normalizer.Form.NFD);
        fileName = fileName.replaceAll("[^\\p{ASCII}]", "");
        return fileName;
    }

    private Path getDirectoryPath(FileDM fileDM) {
        Long groupId;

        if (fileDM.getGroup() == null) {
            groupId = 0L;
        } else {
            groupId = fileDM.getGroup().getId();
        }

        return Paths.get(
                BASE_URL,
                String.valueOf(fileDM.getCategory().getId()),
                String.valueOf(groupId));
    }

    public Path getFilePath(FileDM fileDM) {
        return Paths.get(getDirectoryPath(fileDM).toString(), getFileNameForPath(fileDM));
    }

    private void createDirectoryIfNotExists(Path path) {
        try {
            if (Files.notExists(path)) {
                log.debug("Creating parent directory: " + path.toString());
                Files.createDirectories(path);
            }
        } catch (IOException e) {
            log.error(String.valueOf(e));
            throw new InternalServerErrorException("Vyskytla se neočekávaná chyba na serveru. Obraťte se na svého administrátora.", e);
        }
    }

    /**
     * Building InputStreamResource (spring)
     *
     * @param fileId
     * @return InputStreamResource
     * @throws InternalServerErrorException
     */
    public void downloadFile(Long fileId, OutputStream outputStream) {
        log.debug("Downloading file by id: " + fileId);
        FileDM fileDM = fileService.getFileDetails(fileId);

        this.downloadByIO(fileDM, outputStream);
    }

    public void downloadByIO(FileDM fileDM, OutputStream outputStream) {
        try (BufferedInputStream in = new BufferedInputStream(this.getFileInputStream(fileDM))) {
            byte dataBuffer[] = new byte[1024];
            int bytesRead;
            while ((bytesRead = in.read(dataBuffer, 0, 1024)) != -1) {
                outputStream.write(dataBuffer, 0, bytesRead);
            }
        } catch (IOException e) {
            throw new InternalServerErrorException("Error in downloading...");
        }
    }
}


