package cz.neit.dm.mail;


import cz.neit.dm.entity.Link;
import cz.neit.dm.entity.PasswordResetToken;
import cz.neit.dm.entity.Recipient;
import cz.neit.dm.entity.User;
import cz.neit.dm.exceptions.InvalidStateException;
import cz.neit.dm.properties.MailProperties;
import cz.neit.dm.services.SideLinkService;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.mail.javamail.JavaMailSender;
import org.springframework.mail.javamail.MimeMessageHelper;
import org.springframework.stereotype.Service;
import org.thymeleaf.context.Context;
import org.thymeleaf.spring5.SpringTemplateEngine;

import javax.mail.MessagingException;
import javax.mail.internet.MimeMessage;
import java.nio.charset.StandardCharsets;
import java.util.List;

/**
 * Created by jhruby on 16.10.2018
 */

@Slf4j
@Service
public class EmailServiceImpl implements EmailService {

    @Value("${app.web.downloadPath}")
    private String baseDownloadUrl;

    @Value("${app.mail.baseUrl}")
    private String baseUrl;

    @Value("${app.mail.frontendUrl}")
    private String frontendUrl;

    @Value("${app.mail.from.address}")
    private String fromEmail;

    private final JavaMailSender emailSender;
    private final SideLinkService sideLinkService;
    private final SpringTemplateEngine springTemplateEngine;
    private final MailProperties mailProperties;

    @Autowired
    public EmailServiceImpl(JavaMailSender emailSender, SideLinkService sideLinkService, SpringTemplateEngine springTemplateEngine, MailProperties mailProperties) {
        this.emailSender = emailSender;
        this.sideLinkService = sideLinkService;
        this.springTemplateEngine = springTemplateEngine;
        this.mailProperties = mailProperties;
    }

    //MIME Message
    @Override
    public void sendMimeMessage(Link link, List<Recipient> recipients) {
        try {
            MimeMessage message = emailSender.createMimeMessage();
            MimeMessageHelper helper = new MimeMessageHelper(message,
                    MimeMessageHelper.MULTIPART_MODE_MIXED_RELATED,
                    StandardCharsets.UTF_8.name());

            Context context = new Context();
            context.setVariable("baseUrl", mailProperties.getBaseUrl());
            context.setVariable("companyUrl", mailProperties.getCompanyUrl());
            context.setVariable("facebookUrl", mailProperties.getFacebookUrl());
            context.setVariable("linkedinUrl", mailProperties.getLinkedinUrl());
            context.setVariable("fileName", link.getFile().getFilename());
            context.setVariable("validUntil", link.getValidUntil());
            context.setVariable("downloadLimit", link.getDownloadsLimit());
            context.setVariable("mailText", link.getMailText());
            context.setVariable("subject", link.getMailSubject());

            recipients.forEach(recipient -> {
                log.debug("Sending email to: " + recipient.getEmail());

                context.setVariable("downloadUrl", baseDownloadUrl + sideLinkService.getSideLinkByLinkIdAndRecipientId(
                        link.getId(), recipient.getId()).getToken());

                String html = springTemplateEngine.process("emailTemplate", context);

                try {
                    helper.setTo(recipient.getEmail());
                    helper.setText(html, true);
                    helper.setSubject(link.getMailSubject());
                    helper.setFrom(fromEmail);
                } catch (MessagingException e) {
                    log.error(e.getMessage());
                }

                emailSender.send(message);
                log.debug("Email to '" + recipient.getEmail() + "' sent");
            });
        } catch (Exception e){
            log.error(String.valueOf(e));
            throw new InvalidStateException("Error during sending emails.", e);
        }
    }

    @Override
    public void sendNotification(Link link, List<Recipient> recipients) {
        try {
            MimeMessage message = emailSender.createMimeMessage();
            MimeMessageHelper helper = new MimeMessageHelper(message,
                    MimeMessageHelper.MULTIPART_MODE_MIXED_RELATED,
                    StandardCharsets.UTF_8.name());

            Context context = new Context();
            context.setVariable("baseUrl", mailProperties.getBaseUrl());
            context.setVariable("companyUrl", mailProperties.getCompanyUrl());
            context.setVariable("facebookUrl", mailProperties.getFacebookUrl());
            context.setVariable("linkedinUrl", mailProperties.getLinkedinUrl());
            context.setVariable("fileName", link.getFile().getFilename());
            context.setVariable("validUntil", link.getValidUntil());
            context.setVariable("mailText", "Platnost Vašeho odkazu vypršela.");
            context.setVariable("subject", link.getMailSubject());

            recipients.forEach(recipient -> {
                log.debug("Sending notification email to: " + recipient.getEmail());

                context.setVariable("downloadUrl", baseDownloadUrl + sideLinkService.getSideLinkByLinkIdAndRecipientId(
                        link.getId(), recipient.getId()).getToken());

                String html = springTemplateEngine.process("emailNotificationTemplate", context);

                try {
                    helper.setTo(recipient.getEmail());
                    helper.setText(html, true);
                    helper.setSubject("Zneplatnění - " + link.getMailSubject());
                    helper.setFrom(fromEmail);
                } catch (MessagingException e) {
                    log.error(e.getMessage());
                }

                emailSender.send(message);
                log.debug("Email to '" + recipient.getEmail() + "' sent");
            });
        } catch (Exception e){
            throw new InvalidStateException("Error during sending emails.", e);
        }

    }

    @Override
    public void sendResetTokenEmail(User user, PasswordResetToken passwordResetToken) {
        try {
            MimeMessage message = emailSender.createMimeMessage();
            MimeMessageHelper helper = new MimeMessageHelper(message,
                    MimeMessageHelper.MULTIPART_MODE_MIXED_RELATED,
                    StandardCharsets.UTF_8.name());

            Context context = new Context();
            context.setVariable("baseUrl", mailProperties.getBaseUrl());
            context.setVariable("companyUrl", mailProperties.getCompanyUrl());
            context.setVariable("facebookUrl", mailProperties.getFacebookUrl());
            context.setVariable("linkedinUrl", mailProperties.getLinkedinUrl());
            context.setVariable("mailText", "Uživatel '" + user.getUsername()+ "' si vyžádal změnu hesla. Pokud jste to nebyli vy, tento email ignorujte.");
            context.setVariable("subject", "Resetování hesla");
            context.setVariable("linkExpiration", passwordResetToken.getExpiryDate());
            context.setVariable("passwdResetUrl",mailProperties.getFrontendUrl() + "password/change/" + passwordResetToken.getToken());

            String html = springTemplateEngine.process("emailPasswordResetTemplate", context);

            try {
                helper.setTo(user.getEmail());
                helper.setText(html, true);
                helper.setSubject("Resetování hesla");
                helper.setFrom(fromEmail);
            } catch (MessagingException e) {
                log.error(e.getMessage());
            }

            emailSender.send(message);
            log.debug("Email for reseting password was send to: " + user.getEmail());

        } catch (Exception e) {
            log.error(String.valueOf(e));
            throw new InvalidStateException("Error during sending emails.", e);
        }
    }


    
}
