package cz.neit.dm.security.acl;

import cz.neit.dm.repositories.*;
import cz.neit.dm.security.acl.conditions.FileConditions;
import cz.neit.dm.security.acl.conditions.GroupConditions;
import cz.neit.dm.security.acl.conditions.LinkConditions;
import cz.neit.security.acl.IPermissionsService;
import cz.neit.security.acl.IResourceRepository;
import cz.neit.security.acl.Role;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.util.HashMap;
import java.util.Map;

import static cz.neit.security.acl.conditions.ConditionsFactory.or;

/**
 * Created by jhruby on 12.11.2018
 */
@Service
public class PermissionsService implements IPermissionsService {

    private Map<String, Role> roles = new HashMap<>();
    private Map<String, IResourceRepository> resources = new HashMap<>();

    /**
     * Default constructor which initialize all user roles used within
     * application and assign permission rules to them.
     * @param groupRepository
     * @param fileRepository
     * @param linkRepository
     * @param userRepository
     */
    @Autowired
    public PermissionsService(
            GroupRepository groupRepository,
            FileRepository fileRepository,
            LinkRepository linkRepository,
            UserRepository userRepository
    ) {
        Role user = new Role(Roles.USER);
        Role admin = new Role(Roles.ADMINISTRATOR, user);

        // Endpoints without ID cannot have a conditions

        //*****GROUP
        user
                .addPermissionRules(true, "group", new String[]{"showGroups"})
                .addPermissionRules(true, "group", new String[]{"viewDetail"},
                        GroupConditions::isMember)
                .addPermissionRules(true, "group", new String[]{ "update", "assignUser", "unassignUser", "showUnassigned", "showManagers"},
                        GroupConditions::isManager)

                .addPermissionRules(true, "fileDM", new String[]{"showFiles", "create"})
                .addPermissionRules(true, "fileDM", new String[]{"download"},
                        FileConditions::canDownload)
                .addPermissionRules(true, "fileDM", new String[]{"viewDetail", "isShared"},
                        or(
                                FileConditions::isOwner,
                                FileConditions::isMemberOfGroup
                        ))
                .addPermissionRules(true, "fileDM", new String[]{"update", "delete", "isDeletable"},
                        or(
                                FileConditions::isOwner,
                                FileConditions::isManager
                        ))

                .addPermissionRules(true, "link", new String[]{"showLinks", "create"})
                .addPermissionRules(true, "link", new String[]{"viewDetail", "download"},
                        or(
                                LinkConditions::isOwner,
                                LinkConditions::isMemberOfGroup
                        ))
                .addPermissionRules(true, "link", new String[]{"update", "delete", "shareLink", "disableLink", "resendEmail"},
                        or(
                                LinkConditions::isOwner,
                                LinkConditions::isManager
                        ))
                .addPermissionRules(true, "user", new String[]{"viewUsersInGroup","updateCurrentUser"})
                .addPermissionRules(true,"category", new String[]{"viewAll"})
                .addPermissionRules(true, "history", new String[]{"showHistory"})
                .addPermissionRules(true, "recipient", new String[]{"viewAll"})
        ;


        admin
                .addPermissionRules(true, "group",
                        new String[]{"create", "delete", "update", "showGroups","viewAll", "viewDetail",
                                "isDeletable", "assignManager", "unassignManager", "unassignManagerAndKick",
                                "assignUser", "unassignUser", "showUnassigned", "changeAuthor", "showManagers"})
                .addPermissionRules(true, "fileDM",
                        new String[]{"create", "delete", "update", "showFiles", "viewAll", "viewDetail",
                                "isShared", "isDeletable", "download", "changeAuthor"})
                .addPermissionRules(true, "link",
                        new String[]{"create", "delete", "update", "showLinks", "viewAll", "viewDetail",
                                "download", "shareLink", "disableLink", "resendEmail"})
                .addPermissionRules(true, "user",
                        new String[]{"viewAll","viewDetail","create","updateOther","delete"})
                .addPermissionRules(true, "history", new String[]{"viewAll"})
        ;

        roles.put(user.getName(), user);
        roles.put(admin.getName(), admin);

        // repositories which will be used to find resources by identification
        resources.put("group", groupRepository);
        resources.put("fileDM", fileRepository);
        resources.put("link", linkRepository);
        resources.put("user", userRepository);
    }

    public boolean roleExists(String role) {
        return roles.containsKey(role);
    }

    public Role getRole(String roleString) {
        Role role = roles.get(roleString);
        if (role == null) {
            throw new RuntimeException("Role '" + roleString + "' not found");
        }

        return role;
    }

    public IResourceRepository getResource(String resource) {
        IResourceRepository repository = resources.get(resource);
        if (repository == null) {
            throw new RuntimeException("Resource '" + resource + "' not found");
        }

        return repository;
    }
}
