package cz.neit.dm.services;

import cz.neit.dm.dtos.mappers.DefaultUserSettingsMapper;
import cz.neit.dm.dtos.mappers.DeleteDateMapper;
import cz.neit.dm.dtos.mappers.GroupMapper;
import cz.neit.dm.dtos.modeldto.DefaultUserSettingsDTO;
import cz.neit.dm.dtos.modeldto.UserDTO;
import cz.neit.dm.dtos.modeldto.UserSimpleDTO;
import cz.neit.dm.entity.DefaultUserSettings;
import cz.neit.dm.entity.User;
import cz.neit.dm.exceptions.BadRequestException;
import cz.neit.dm.exceptions.ResourceNotFoundException;
import cz.neit.dm.repositories.DefaultUserSettingsRepository;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.security.crypto.bcrypt.BCrypt;
import org.springframework.security.crypto.bcrypt.BCryptPasswordEncoder;
import org.springframework.stereotype.Service;

/**
 * Created by jhruby on 16.04.2019
 */
@Slf4j
@Service
@RequiredArgsConstructor
public class DefaultUserSettingsServiceImpl implements DefaultUserSettingsService {

    private final UserService userService;
    private final BCryptPasswordEncoder bCryptPasswordEncoder;
    private final DefaultUserSettingsMapper defaultUserSettingsMapper;
    private final GroupMapper groupMapper;
    private final DeleteDateMapper deleteDateMapper;
    private final DefaultUserSettingsRepository defaultRepository;
    private final CategoryService categoryService;

    @Override
    public DefaultUserSettings getDefaultUserSettingsDetails(Long defId) {
        return defaultRepository.findById(defId)
                .orElseThrow(() -> new ResourceNotFoundException("Defaultní nastavení uživatele nebylo nalezeno"));
    }

    @Override
    public DefaultUserSettingsDTO getDefaultUserSettingsDetailsByCurrentUser(User currentUser) {

        if (currentUser.getDefaultUserSettings() != null) {
            DefaultUserSettings defaultUser = currentUser.getDefaultUserSettings();
            log.debug("Getting default user settings for user '" + currentUser.getFullname() + "' ");
            return defaultUserSettingsMapper.defaultSetToDefaultSetDTO(defaultUser);
        } else {
            return null;
        }

    }

    @Override
    public UserSimpleDTO updateCurrentUser(Long id, UserDTO userDTO) {
        User user = userService.getUserDetail(id);
        user.setEmail(userDTO.getEmail());
        user.setFirstname(userDTO.getFirstname());
        user.setLastname(userDTO.getLastname());

        if (userDTO.getPassword() != null && userDTO.getOldPassword() != null) {
            if (BCrypt.checkpw(userDTO.getOldPassword(), user.getPassword())) {
                user.setOldPassword(user.getPassword());
                user.setPassword(bCryptPasswordEncoder.encode(userDTO.getPassword()));
            } else {
                throw new BadRequestException("Zadané původní heslo neodpovídá.");
            }
        }

        log.debug("Update current user");
        return userService.saveUserAndReturnDTO(user);
    }

    @Override
    public void updateUserPassword(String newPassword, User user) {
        user.setPassword(bCryptPasswordEncoder.encode(newPassword));
        userService.simpleSaveUser(user);
        log.debug("Password for user: '" + user.getFullname() + "' was changed.");
    }

    @Override
    public DefaultUserSettingsDTO setDefaultUserSettings(Long userId, DefaultUserSettingsDTO dto) {

        User user = userService.getUserDetail(userId);
        DefaultUserSettings defaultUserSettings;

        if (user.getDefaultUserSettings() != null) {
            defaultUserSettings = this.getDefaultUserSettingsDetails(user.getDefaultUserSettings().getId());
        } else {
            defaultUserSettings = new DefaultUserSettings();
        }

        defaultUserSettings.setCategory(categoryService.getCategoryByName(dto.getCategory()));
        defaultUserSettings.setGroup(groupMapper.simpleGroupDtoToGroup(dto.getGroup()));
        defaultUserSettings.setDeleteDate(deleteDateMapper.deleteDtoToDelete(dto.getDeleteDate()));
        defaultUserSettings.setOnlyAuthorCanDownload(dto.isOnlyAuthorCanDownload());
        defaultUserSettings.setMailSubject(dto.getMailSubject());
        defaultUserSettings.setMailText(dto.getMailText());
        defaultUserSettings.setValidUntil(dto.getValidUntil());
        defaultUserSettings.setDownloadsLimit(dto.getDownloadsLimit());

        user.setDefaultUserSettings(defaultRepository.save(defaultUserSettings));
        userService.simpleSaveUser(user);

        return defaultUserSettingsMapper.defaultSetToDefaultSetDTO(defaultUserSettings);
    }


}
