package cz.neit.dm.services;

import cz.neit.dm.dtos.mappers.FileMapper;
import cz.neit.dm.dtos.modeldto.FileDTO;
import cz.neit.dm.dtos.modeldto.FileSimpleDTO;
import cz.neit.dm.dtos.modeldto.DownloadTokenDTO;
import cz.neit.dm.entity.FileDM;
import cz.neit.dm.entity.Link;
import cz.neit.dm.exceptions.ResourceNotFoundException;
import cz.neit.dm.repositories.FileRepository;
import cz.neit.dm.security.SecurityContextUtils;
import cz.neit.security.acl.AuthorizatorService;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.io.FilenameUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Lazy;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.multipart.MultipartFile;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.time.LocalDate;
import java.util.*;
import java.util.concurrent.atomic.AtomicReference;
import java.util.stream.Collectors;

/**
 * Created by jhruby on 26.08.2018
 */

@Slf4j
@Service
@Transactional
public class FileServiceImpl implements FileService {

    private final FileRepository fileRepository;
    private final FileMapper fileMapper;
    private final CategoryService categoryService;
    private final GroupMembershipService groupMembershipService;
    private final DeleteDateService deleteDateService;
    private final AuthorizatorService authorizatorService;

    private Map<String, FileDM> tokensForFile = new HashMap<>();

    private LinkService linkService;

    @Autowired
    public FileServiceImpl(FileRepository fileRepository, FileMapper fileMapper, CategoryService categoryService, GroupMembershipService groupMembershipService, DeleteDateService deleteDateService, AuthorizatorService authorizatorService) {
        this.fileRepository = fileRepository;
        this.fileMapper = fileMapper;
        this.categoryService = categoryService;
        this.groupMembershipService = groupMembershipService;
        this.deleteDateService = deleteDateService;
        this.authorizatorService = authorizatorService;
    }

    @Lazy
    @Autowired
    public void setLinkService(LinkService linkService) {
        this.linkService = linkService;
    }

    // *********** GETTERS LIST
    @Override
    public List<FileDM> getAllFiles() {
        List<FileDM> list = new ArrayList<>();
        fileRepository.findAll().forEach(list::add);
        return list;
    }

    @Override
    public List<FileDTO> getAllFilesDTO() {
        return getAllFiles().stream()
                .map(fileMapper::fileDmToFileDTO)
                .collect(Collectors.toList());
    }

    @Override
    public List<FileDM> getFilesByAuthorId(Long userId) {
        return fileRepository.findAllByAuthorId(userId);
    }

    @Override
    public List<FileDTO> getFilesDTOByAuthorId(Long userId) {
        return getFilesByAuthorId(userId).stream()
                .map(fileMapper::fileDmToFileDTO)
                .collect(Collectors.toList());
    }

    @Override
    public List<FileDM> getFilesByUserIdWithoutGroup(Long userId) {
        log.debug("Getting all files by AuthorId without group");

        return fileRepository.findAllByGroupIsNullAndAuthorId(userId);
    }

    @Override
    public List<FileDM> getFilesByUserIdAndGroupId(Long userId, Long groupId) {
        return fileRepository.findAllByGroupIdAndAuthorId(groupId, userId);
    }


    @Override
    public List<FileDTO> getAllFilesForCurrentUser() {
        if (authorizatorService.isAllowed("fileDM", "viewAll")) {
            log.debug("Getting All FilesDTO for administrator");
            return this.getAllFilesDTO();
        } else {
            log.debug("Getting All FilesDTO for user");

            return fileRepository.findAllByGroupIdIn(groupMembershipService.getGroupIdsForCurrentUser())
                    .stream()
                    .map(fileMapper::fileDmToFileDTO)
                    .collect(Collectors.toCollection(this::getFilesWithoutGroup));
        }
    }

    @Override
    public List<Long> getAllFileIdsForCurrentUser() {
        return getAllFilesForCurrentUser().stream()
                .map(FileDTO::getId)
                .collect(Collectors.toList());
    }

    @Override
    public List<FileDTO> getFilesWithoutGroup() {
        log.debug("Getting all Files without group for current user.");

        return fileRepository.findAllByGroupIsNull().stream()
                .filter(fileDM -> fileDM.getAuthor().getId().equals(SecurityContextUtils.getCurrentLoggedUser().getId()))
                .map(fileMapper::fileDmToFileDTO)
                .collect(Collectors.toList());
    }

    @Override
    public List<FileDM> getFilesByGroup(Long groupId) {
        return fileRepository.findAllByGroupId(groupId);
    }


    // *********** GETTERS
    @Override
    public FileDM getFileDetails(Long fileId) {
        return fileRepository.findById(fileId)
                .orElseThrow(() -> new ResourceNotFoundException("Soubor nebyl nalezen."));
    }

    @Override
    public FileDTO getFileDetailsDTO(Long fileId) {
        return fileMapper.fileDmToFileDTO(fileRepository.findById(fileId)
                .orElseThrow(() -> new ResourceNotFoundException("Soubor nebyl nalezen.")));
    }

    @Override
    public FileDM getFileByTokenFromAPP(String token) {

        return tokensForFile.entrySet()
                .stream()
                .filter(stringFileDMEntry -> stringFileDMEntry.getKey().equals(token))
                .map(Map.Entry::getValue)
                .findFirst().orElseThrow(() -> new ResourceNotFoundException("Token pro soubor neexistuje nebo nebyl nalezen."));
    }

    @Override
    public FileDTO getFileDetailsWithDescription(Long fileId) {
        FileDTO fileDTO = getFileDetailsDTO(fileId);
        fileDTO.getLinks().forEach(link -> {
            String description = link.getDescription();
            int maxLength = (description.length() < 50) ? description.length() : 50;

            if (maxLength == 50) {
                description = description.substring(0, maxLength) + "...";
            }
            link.setDescription(description);
        });

        fileDTO.setDeleteDateSimple(deleteDateService.getDateWhenFileWillBeDeleted(fileMapper.fileDtoToFileDM(fileDTO)));
        return fileDTO;
    }

    @Override
    public DownloadTokenDTO getDownloadTokenForFile(Long fileId) {
        //TODO ukládat spíš do DBS, ted musí appka bežet v jedné instanci
        DownloadTokenDTO token = new DownloadTokenDTO();
        token.setValue(UUID.randomUUID().toString());

        tokensForFile.put(token.getValue(), getFileDetails(fileId));

        log.debug(token.getValue());
        return token;
    }


    // *********** CRUD
    @Override
    public FileDM createFileWithoutLink(FileDTO fileDTO, MultipartFile multipartFile) {
        log.debug("Saving file " + fileDTO.getFilename() + " to database.");
        FileDM fileDM = fileMapper.fileDtoToFileDM(fileDTO);

        fileDM.setFilesize(multipartFile.getSize());
        fileDM.setFormat(FilenameUtils.getExtension(fileDTO.getFilename()));
        fileDM.setUploadedAt(LocalDate.now());
        fileDM.setAuthor(SecurityContextUtils.getCurrentLoggedUser());
        fileDM.setCategory(categoryService.createIfCategoryNotExistsByName(fileDTO.getCategory()));

        return fileRepository.save(fileDM);
    }

    @Override
    public FileDTO updateFile(FileDTO fileDTO, Long fileId) {
        log.debug("Updating file on id: " + fileId);

        FileDM fileDM = getFileDetails(fileId);
        fileDM.setDescription(fileDTO.getDescription());
        fileDM.setDeleteDate(deleteDateService.updateDeleteDate(fileDTO));
        fileDM.setOnlyAuthorCanDownload(fileDTO.isOnlyAuthorCanDownload());

        return fileMapper.fileDmToFileDTO(fileRepository.save(fileDM));
    }

    @Override
    public void deleteFile(Long fileId, Path path) {
        log.debug("Delete file by id: " + fileId + " from database.");
        FileDM fileDM = getFileDetails(fileId);

        try {
            fileDM.getDownloadHistory().forEach(downloadHistory -> {
                downloadHistory.setSideLink(null);
                downloadHistory.setLink(null);
                downloadHistory.setFile(null);
            });

            fileRepository.deleteById(fileId);
            Files.delete(path);
        } catch (IOException e) {
        }
    }


    // *********** CONDITIONS FOR ENDPOINTS

    @Override
    public boolean isShared(Long fileId) {
        Optional sharedLink = getFileDetails(fileId).getLinks().stream()
                .filter(Link::isShared)
                .findFirst();

        return sharedLink.isPresent();
    }

    @Override
    public boolean isAuthorHasFilesInGroup(Long userId, Long groupId) {
        return getFilesByUserIdAndGroupId(userId, groupId).isEmpty();
    }

    @Override
    public boolean isDownloadable(Long fileId) {
        FileDM fileDM = getFileDetails(fileId);

        boolean isAuthor = SecurityContextUtils.getCurrentLoggedUser().getId().equals(fileDM.getAuthor().getId());
        return !fileDM.isOnlyAuthorCanDownload() || isAuthor || authorizatorService.isAllowed("fileDM", "download");
    }

    @Override
    public boolean isAllLimitsExceeded(Long fileId, Link link) {
        if (linkService.isDownloadsCountLimitExceeded(link)) {
            linkService.disableLink(link.getId());
        }
        return !isShared(fileId);
    }

    // *********** SUPPORTIVE METHODS FOR PROPERTIES
    @Override
    public List<FileSimpleDTO> mapFilesByGroupToSimpleDTO(List<FileDM> fileDMList) {
        return fileDMList.stream()
                .map(fileMapper::fileDmToSimpleFileDTO)
                .collect(Collectors.toList());
    }

    @Override
    public void setLatestActiveLinkDate(FileDM file) {
        AtomicReference<LocalDate> sharedUpTo = new AtomicReference<>(null);

        Optional<Link> link = file.getLinks()
                .stream()
                .findFirst();

        if (link.isPresent()) {
            sharedUpTo.set(link.get().getValidUntil());

            file.getLinks()
                    .stream()
//                    .filter(Link::isShared)
                    .map(Link::getValidUntil)
                    .forEach(localDate -> {
                        if (localDate.isAfter(sharedUpTo.get())) {
                            sharedUpTo.set(localDate);
                        }
                    });

            log.debug("Set sharedUpTo parameter.");
            file.setSharedUpTo(sharedUpTo.get());

        } else {
            log.debug("No active Links for File.");
            file.setSharedUpTo(null);
        }
        fileRepository.save(file);
    }

    @Override
    public void setNewLastDownloadViaLimit(Long fileId) {
        log.debug("Set last download via limit");
        FileDM fileDM = getFileDetails(fileId);

        fileDM.setLastDownloadViaLink(LocalDate.now());
        fileRepository.save(fileDM);
    }

    @Override
    public void addNewLinkToFileDM(Link link) {
        FileDM fileDM = getFileDetails(link.getFile().getId());
        List<Link> linkList = fileDM.getLinks();
        linkList.add(link);
        fileDM.setLinks(linkList);

        fileRepository.save(fileDM);
    }

    @Override
    public void save(FileDM fileDM) {
        fileRepository.save(fileDM);
    }


    // *********** SCHEDULING


    @Override
    public List<FileDM> getDeprecatedFiles() {
        return getAllFiles()
                .stream()
                .filter(deleteDateService::isFileToDelete)
                .filter(fileDM -> !isShared(fileDM.getId()))
                .collect(Collectors.toList());
    }


    /**
     * Method for getting confirmation about date conflict between deleteDate and link.validUntil
     * Return true, if some conflict exist (makes no differents if link is shared
     *
     * @param fileDM
     * @return
     */
    @Override
    public boolean isAnyFileLinkIncorrect(FileDM fileDM) {
        if (fileDM.getDeleteDate().getWillBeDeletedAt() != null) {
            return fileDM.getLinks()
                    .stream()
                    .anyMatch(link -> fileDM.getDeleteDate().getWillBeDeletedAt().isBefore(link.getValidUntil()));
        } else {
            return false;
        }
    }


}
