package cz.neit.dm.services;

import cz.neit.dm.dtos.mappers.UserMapper;
import cz.neit.dm.dtos.modeldto.UserSimpleDTO;
import cz.neit.dm.entity.FileDM;
import cz.neit.dm.entity.Group;
import cz.neit.dm.entity.GroupMembership;
import cz.neit.dm.entity.User;
import cz.neit.dm.entity.enums.MembershipType;
import cz.neit.dm.exceptions.BadRequestException;
import cz.neit.dm.exceptions.ResourceNotFoundException;
import cz.neit.dm.repositories.FileRepository;
import cz.neit.dm.repositories.GroupMembershipRepository;
import cz.neit.dm.repositories.GroupRepository;
import cz.neit.dm.repositories.UserRepository;
import cz.neit.dm.security.SecurityContextUtils;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Service;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

import static cz.neit.dm.entity.enums.MembershipType.MANAGER;
import static cz.neit.dm.entity.enums.MembershipType.USER;

/**
 * Created by jhruby on 25.07.2018
 */

@Slf4j
@Service
@RequiredArgsConstructor
public class GroupMembershipServiceImpl implements GroupMembershipService {

    private final GroupMembershipRepository groupMembershipRepository;
    private final GroupRepository groupRepository;
    private final UserRepository userRepository;
    private final FileRepository fileRepository;
    private final UserMapper userMapper;

    // *********** GETTERS
    @Override
    public List<GroupMembership> getMembershipByUserId(Long userId) {
        log.debug("Getting list of memberships by UserId: " + userId);
        return groupMembershipRepository.findAllByUserId(userId);
    }

    @Override
    public List<GroupMembership> getManagersForGroup(Long groupId) {
        log.debug("Getting list of group managers by groupId: " + groupId);
        return groupMembershipRepository.findAllByGroupIdAndMembershipType(groupId, MembershipType.MANAGER);
    }

    @Override
    public GroupMembership getMembershipByGroupIdAndUserId(Long groupId, Long userId) {
        log.debug("Get GroupMembership by group ID: " + groupId + " and user ID: " + userId);
        return groupMembershipRepository.findByGroupIdAndUserId(groupId,userId)
                .orElseThrow(() -> new ResourceNotFoundException("Neexistuje požadované spojení mezi tabulkou a uživatelem"));
    }


    @Override
    public List<Long> getGroupIdsForCurrentUser() {
        return getMembershipByUserId(SecurityContextUtils.getCurrentLoggedUser().getId())
                .stream()
                .map(groupMembership -> groupMembership.getGroup().getId())
                .collect(Collectors.toList());
    }

    // *********** DELETE

    @Override
    public void deleteManagerGroupMembership(Long groupId, Long userId) {
        log.debug("Deleting GroupMembership by group ID: +" + groupId + " and user ID:" + userId);
        groupMembershipRepository.deleteById(getMembershipByGroupIdAndUserId(groupId,userId).getId());
    }

    @Override
    public void deleteGroupMembership(Long groupId, Long userId, User newAuthor) {
        log.debug("Deleting GroupMembership by group ID: " + groupId + " and user ID: " + userId);

        List<FileDM> filesByGroupAndAuthor = fileRepository.findAllByGroupIdAndAuthorId(groupId, userId);

        if (!filesByGroupAndAuthor.isEmpty()) {
            setManagerAsAAuthor(filesByGroupAndAuthor, newAuthor);
        }
        groupMembershipRepository.deleteById(getMembershipByGroupIdAndUserId(groupId,userId).getId());
    }


    @Override
    public boolean hasUserFilesInGroup(Long groupId, Long userId) {
        List<FileDM> fileDMList = fileRepository.findAllByGroupIdAndAuthorId(groupId, userId);
        return !fileDMList.isEmpty();

    }

    private void setManagerAsAAuthor(List<FileDM> files, User newAuthor) {
        log.debug("Setting new authors");
        files.forEach(fileDM -> {
            fileDM.getLinks().forEach(link -> {
                if (link.getAuthor().getUsername().equals(fileDM.getAuthor().getUsername())) {
                    link.setAuthor(newAuthor);
                }
            });
            fileDM.setAuthor(newAuthor);
            fileRepository.save(fileDM);
            log.debug("File with ID: " + fileDM.getId() + " has new author, it's groupManager with name: " + newAuthor.getFullname());
        });
    }

    @Override
    public List<UserSimpleDTO> getManagersOfGroupForNewAuthor(Long groupId, Long userId) {
        List<UserSimpleDTO> list = getManagersForGroup(groupId)
                .stream()
                .filter(groupMembership -> !groupMembership.getUser().getId().equals(userId))
                .map(GroupMembership::getUser)
                .map(userMapper::userToSimpleUserDTO)
                .collect(Collectors.toList());
        if (!list.isEmpty()) {
            return list;
        } else {
            throw new BadRequestException("Neexistuje správce, kterému by bylo vlastnictví souboru předáno");
        }
    }

    @Override
    public List<UserSimpleDTO> getAllManagers() {
        log.debug("Getting all possible managers");
        List<Group> groups = new ArrayList<>();
        groupRepository.findAll().forEach(groups::add);
//
        List<UserSimpleDTO> users = new ArrayList<>();
        groups.forEach(group -> users.addAll(getManagersForGroup(group.getId())
                .stream()
                .map(GroupMembership::getUser)
                .map(userMapper::userToSimpleUserDTO)
                .collect(Collectors.toList())));

        return users.stream().distinct().collect(Collectors.toList());
    }


    // *********** ASSIGN USERS and MANAGERS to GROUP
    @Override
    public GroupMembership assignUserToGroup(Long groupId, Long userId) {
        log.debug("Adding USER to group, group ID: " + groupId + " and user ID: " + userId);
        return saveMembershipToRepository(groupId, userId, USER);
    }

    @Override
    public GroupMembership assignManagerToGroup(Long groupId, Long userId) {
        log.debug("Adding MANAGER to group, group ID: " + groupId + " and user ID: " + userId);
        return saveMembershipToRepository(groupId,userId,MANAGER);
    }

    // *********** PRIVATE - SAVE
    private GroupMembership saveMembershipToRepository(Long groupId, Long userId, MembershipType membershipType) {
        //Not use services here, due to cross dependency of FileServiceImpl
        User user = userRepository.findById(userId).orElseThrow(() -> new ResourceNotFoundException("User not exists."));
        Group group = groupRepository.findById(groupId).orElseThrow(() -> new ResourceNotFoundException("Group not exists."));

        Optional<GroupMembership> membership = groupMembershipRepository.findByGroupIdAndUserId(groupId,userId);
        GroupMembership groupMembership;

        if (membership.isPresent()) {
            groupMembership = membership.get();
            if(groupMembership.getMembershipType() == membershipType) {
                throw new BadRequestException("Uživatel '" + user.getUsername() + "' je již členem ve skupině '" + group.getName() + "' na pozici " + membershipType);
            }
            groupMembership.setMembershipType(membershipType);
        } else {
            groupMembership = new GroupMembership(user,group,membershipType);
        }
        return groupMembershipRepository.save(groupMembership);
    }
}
