package cz.neit.dm.services;

import cz.neit.dm.dtos.mappers.GroupMapper;
import cz.neit.dm.dtos.mappers.UserMapper;
import cz.neit.dm.dtos.modeldto.GroupDTO;
import cz.neit.dm.dtos.modeldto.GroupSimpleDTO;
import cz.neit.dm.dtos.modeldto.GroupUsersDTO;
import cz.neit.dm.dtos.modeldto.UserSimpleDTO;
import cz.neit.dm.entity.Group;
import cz.neit.dm.entity.GroupMembership;
import cz.neit.dm.exceptions.BadRequestException;
import cz.neit.dm.exceptions.ResourceNotFoundException;
import cz.neit.dm.repositories.GroupMembershipRepository;
import cz.neit.dm.repositories.GroupRepository;
import cz.neit.dm.security.SecurityContextUtils;
import cz.neit.security.acl.AuthorizatorService;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Service;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

import static cz.neit.dm.entity.enums.MembershipType.MANAGER;
import static cz.neit.dm.entity.enums.MembershipType.USER;

/**
 * Created by jhruby on 25.07.2018
 */

@Slf4j
@Service
@RequiredArgsConstructor
public class GroupServiceImpl implements GroupService {

    private final GroupRepository groupRepository;
    private final GroupMapper groupMapper;
    private final UserMapper userMapper;
    private final GroupMembershipRepository groupMembershipRepository;
    private final FileService fileService;
    private final AuthorizatorService authorizatorService;

    // *********** GETTERS
    @Override
    public GroupDTO getGroupDetailsDTO(Long id) {
        log.debug("Getting GroupDTO by id: " + id);

        Optional<GroupDTO> groupDTO = groupRepository.findById(id)
                .map(group -> {
                    GroupDTO g = groupMapper.groupToGroupDTO(group);
                    g.setMembers(getGroupUserDTO(group.getId()));
                    g.setFiles(fileService.mapFilesByGroupToSimpleDTO(fileService.getFilesByGroup(id)));
                    return g;
                });
        if (!groupDTO.isPresent()){
            log.error("Skupina s ID " + id + " nebyla nalezena.");
            throw new ResourceNotFoundException("Požadovaná skupina nebyla nalezena.");
        }
        return groupDTO.get();
    }

    @Override
    public Group getGroupDetails(Long id) {
        log.debug("Getting Group by id: " + id);
        return groupRepository.findById(id)
                .orElseThrow(() -> new ResourceNotFoundException("Požadovaná skupina nebyla nalezena."));
    }

    @Override
    public GroupDTO getGroupDetailsByNameDTO(String name) {
        log.debug("Getting Group by name: " + name);

        Optional<GroupDTO> groupDTO = groupRepository.findByName(name)
                .map(group -> {
                    GroupDTO g = groupMapper.groupToGroupDTO(group);
                    g.setMembers(getGroupUserDTO(group.getId()));
                    g.setFiles(fileService.mapFilesByGroupToSimpleDTO(fileService.getFilesByGroup(g.getId())));
                    return g;
                });

        if (!groupDTO.isPresent()){
            log.error("Skupina s jménem " + name + " nebyla nalezena.");
            throw new ResourceNotFoundException("Požadovaná skupina nebyla nalezena.");
        }
        return groupDTO.get();
    }


    // *********** GETTERS LIST
    @Override
    public List<Group> getAllGroups() {
        log.debug("Getting All Groups.");

        List<Group> groups = new ArrayList<>();
        groupRepository.findAll().forEach(groups::add);
        return groups;
    }

    @Override
    public List<GroupDTO> getAllGroupsDTO() {
        log.debug("Getting All GroupsDTO");

        return getAllGroups().stream()
                .map(group -> {
                    GroupDTO groupDTO = groupMapper.groupToGroupDTO(group);
                    groupDTO.setMembers(getGroupUserDTO(group.getId()));
                    return groupDTO;
                })
                .collect(Collectors.toList());
    }

    @Override
    public List<GroupDTO> getAllUnassignedGroups(Long userId) {
        log.debug("Getting All unassigned Groups.");

        List<Long> listOfIds = groupMembershipRepository.findAllByUserId(userId)
                .stream()
                .map(groupMembership -> groupMembership.getGroup().getId())
                .collect(Collectors.toList());

        if (listOfIds.isEmpty()) {
            return getAllGroupsDTO();
        } else {
            return groupRepository.findAllByIdNotIn(listOfIds)
                    .stream()
                    .map(groupMapper::groupToGroupDTO)
                    .collect(Collectors.toList());
        }
    }

    @Override
    public List<GroupDTO> getAllGroupsByCurrentUser() {
        if (authorizatorService.isAllowed("group", "viewAll")) {
            log.debug("Getting All Groups for administrator");
            return getAllGroupsDTO();
        } else {
            log.debug("Getting All Groups for current user");

            return getGroupMembershipForCurrentUser()
                    .stream()
                    .map(GroupMembership::getGroup)
                    .map(group -> {
                        GroupDTO groupDTO = groupMapper.groupToGroupDTO(group);
                        groupDTO.setMembers(getGroupUserDTO(group));
                        return groupDTO;
                    })
                    .collect(Collectors.toList());
        }
    }


    // *********** CRUD
    @Override
    public GroupSimpleDTO updateGroup(Long id, GroupSimpleDTO groupSimpleDTO) {
        log.debug("Updating Group, id: " + id);

        Group group = groupRepository.findById(id).orElseThrow(() -> new ResourceNotFoundException("User on id " + id + "not found."));
        group.setName(groupSimpleDTO.getName());
        group.setDescription(groupSimpleDTO.getDescription());

        return saveAndReturnGroup(group);
    }

    @Override
    public GroupSimpleDTO createGroup(GroupSimpleDTO groupSimpleDTO) {
        log.debug("Creating new Group, name: " + groupSimpleDTO.getName());
        return saveAndReturnGroup(groupMapper.simpleGroupDtoToGroup(groupSimpleDTO));
    }

    @Override
    public void deleteGroup(Long id) {
        if(isDeletable(id)) {
            log.debug("Deleting Group, id: " + id);
            groupRepository.deleteById(id);
        } else {
            throw new BadRequestException("Skupina nemůže být smazána, protože obsahuje nahrané soubory");
        }
    }


    // *********** CONDITIONS FOR ENDPOINTS
    @Override
    public boolean isDeletable(Long groupId) {
        return !getGroupDetailsDTO(groupId).getFiles().stream().findAny().isPresent();
    }


    // *********** PRIVATE - SAVE
    private GroupSimpleDTO saveAndReturnGroup(Group group) {
        return groupMapper.groupToSimpleGroupDTO(groupRepository.save(group));
    }

    private GroupUsersDTO getGroupUserDTO(Long groupId) {
        GroupUsersDTO groupUsersDTO = new GroupUsersDTO();
        List<UserSimpleDTO> users = new ArrayList<>();
        List<UserSimpleDTO> managers = new ArrayList<>();

        groupMembershipRepository.findAllByGroupId(groupId).forEach(
                pair -> {
                    if (pair.getMembershipType().equals(MANAGER)) {
                        managers.add(userMapper.userToSimpleUserDTO(pair.getUser()));
                    } else {
                        users.add(userMapper.userToSimpleUserDTO(pair.getUser()));
                    }
                }
        );

        groupUsersDTO.setUsers(users);
        groupUsersDTO.setManagers(managers);
        return groupUsersDTO;
    }

    private GroupUsersDTO getGroupUserDTO(Group group) {
        GroupUsersDTO groupUsersDTO = new GroupUsersDTO();

        groupUsersDTO.setManagers(
                group.getGroupMembership().stream()
                        .filter(gm -> gm.getMembershipType().equals(MANAGER))
                        .map(GroupMembership::getUser)
                        .map(userMapper::userToSimpleUserDTO)
                        .collect(Collectors.toList())
        );

        group.getGroupMembership().stream()
                .filter(gm -> gm.getMembershipType().equals(USER))
                .map(GroupMembership::getUser)
                .map(userMapper::userToSimpleUserDTO)
                .collect(Collectors.toCollection(groupUsersDTO::getUsers));

        return groupUsersDTO;
    }

    private List<GroupMembership> getGroupMembershipForCurrentUser() {
        return groupMembershipRepository
                .findAllByUserId(SecurityContextUtils.getCurrentLoggedUser().getId());
    }

}
