package cz.neit.dm.services;

import cz.neit.dm.dtos.mappers.LinkMapper;
import cz.neit.dm.dtos.modeldto.DeleteDateSimpleDTO;
import cz.neit.dm.dtos.modeldto.LinkDTO;
import cz.neit.dm.dtos.modeldto.LinkValidationDTO;
import cz.neit.dm.entity.Link;
import cz.neit.dm.entity.Recipient;
import cz.neit.dm.entity.SideLink;
import cz.neit.dm.exceptions.BadRequestException;
import cz.neit.dm.exceptions.ResourceNotFoundException;
import cz.neit.dm.io.LinkManager;
import cz.neit.dm.mail.EmailService;
import cz.neit.dm.repositories.LinkRepository;
import cz.neit.dm.security.SecurityContextUtils;
import cz.neit.security.acl.AuthorizatorService;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.time.LocalDate;
import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;

/**
 * Created by jhruby on 05.09.2018
 */

@Slf4j
@Service
@RequiredArgsConstructor
public class LinkServiceImpl implements LinkService {

    private final LinkRepository linkRepository;
    private final FileService fileService;
    private final LinkMapper linkMapper;
    private final LinkManager linkManager;
    private final SideLinkService sideLinkService;
    private final RecipientService recipientService;
    private final EmailService emailService;
    private final GroupMembershipService groupMembershipService;
    private final AuthorizatorService authorizatorService;
    private final DeleteDateService deleteDateService;

    // *********** GETTERS LIST
    @Override
    public List<Link> getAllLinks() {
        List<Link> linkList = new ArrayList<>();
        linkRepository.findAll().forEach(linkList::add);

        return linkList;
    }

    @Override
    public List<LinkDTO> getAllLinksDTO() {
        return getAllLinks().stream()
                .map(linkMapper::linkToLinkDTO)
                .collect(Collectors.toList());
    }

    @Override
    public List<LinkDTO> getLinksForCurrentUser() {
        if (authorizatorService.isAllowed("link", "viewAll")) {
            log.debug("Getting All Links for administrator");
            return this.getAllLinksDTO();
        } else {
            log.debug("Getting All Links for current user.");

            return linkRepository.findAllByFile_GroupIdIn(groupMembershipService.getGroupIdsForCurrentUser())
                    .stream()
                    .map(linkMapper::linkToLinkDTO)
                    .collect(Collectors.toCollection(this::getLinksDTOWithoutGroup));
        }
    }

    @Override
    public List<LinkDTO> getLinksDTOWithoutGroup() {
        log.debug("Getting All Files without group for current user.");

        return linkRepository.findAllByFile_GroupIsNull()
                .stream()
                .filter(link -> link.getAuthor().getId().equals(SecurityContextUtils.getCurrentLoggedUser().getId()))
                .map(linkMapper::linkToLinkDTO)
                .collect(Collectors.toList());
    }

    @Override
    public List<Recipient> getRecipientsByLinkId(Long linkId) {
        return getLinkDetails(linkId).getSideLinks().stream()
                .map(SideLink::getRecipient)
                .collect(Collectors.toList());
    }


    // *********** GETTERS
    @Override
    public Link getLinkDetails(Long linkId) {
        return linkRepository.findById(linkId)
                .orElseThrow(() -> new ResourceNotFoundException("Odkaz nebyl nalezen."));
    }

    @Override
    public LinkDTO getLinkDetailDtoWithRecipients(Long linkId) {
        LinkDTO linkDTO = linkMapper.linkToLinkDTO(getLinkDetails(linkId));

        List<Recipient> recipients = getRecipientsByLinkId(linkId);

        List<String> emails = recipients.stream().map(Recipient::getEmail).collect(Collectors.toList());

        linkDTO.setRecipients(emails);
        return linkDTO;
    }

    // *********** SHARE OR DISABLE LINK FOR CUSTOMERS
    @Override
    public LinkDTO shareLink(Long linkId) {
        Link link = getLinkDetails(linkId);

        if (link.getValidUntil().isAfter(LocalDate.now())) {
            log.debug("Set share for link on id: " + linkId);
            link.setMainToken(linkManager.getNewUuid4());
            sideLinkService.shareSideLink(linkId);
            emailService.sendMimeMessage(link, this.getRecipientsByLinkId(link.getId()));

            link.setShared(true);
            link.setLastShare(LocalDate.now());
            fileService.setLatestActiveLinkDate(link.getFile());

            saveLinkAndReturnDTO(link);
            return getLinkDetailDtoWithRecipients(linkId);
        } else {
            throw new BadRequestException("Vypršelo datum platnosti odkazu. Není možné odeslat email příjemcům, s prošlým datem platnosti linku.");
        }
    }

    @Override
    public LinkDTO disableLink(Long linkId) {
        log.debug("Disable share for link on id: " + linkId);
        Link link = getLinkDetails(linkId);
        link.setShared(false);
        link.setDownloadsCount(0L);
        link.setMainToken(null);
        sideLinkService.disableSideLink(linkId);
        emailService.sendNotification(link, this.getRecipientsByLinkId(link.getId()));

        fileService.setLatestActiveLinkDate(link.getFile());
        saveLinkAndReturnDTO(link);
        return getLinkDetailDtoWithRecipients(linkId);
    }


    // *********** CRUD
    @Override
    @Transactional
    public LinkDTO createLink(LinkDTO linkDTO) {
        log.debug("Create download link for file ");
        Link link = linkMapper.linkDtoToLink(linkDTO);
        link.setDownloadsCount(0L);
        link.setCreatedAt(LocalDate.now());
        link.setAuthor(SecurityContextUtils.getCurrentLoggedUser());
        link.setShared(false);

        return saveNewLinkAndReturnDTO(link, linkDTO.getRecipients());
    }

    @Override
    public LinkDTO updateLink(Long linkId, LinkDTO linkDTO) {

        Link link = getLinkDetails(linkId);

        if (!link.isShared()) {

            log.debug("Updating link on id: " + linkId);

            link.setMailSubject(linkDTO.getMailSubject());
            link.setMailText(linkDTO.getMailText());
            link.setValidUntil(linkDTO.getValidUntil());
            link.setDownloadsLimit(linkDTO.getDownloadsLimit());
            link.setDescription(linkDTO.getDescription());

            log.debug("All SideLinks with LinkId: " + linkId + " was deleted!");
            sideLinkService.deleteSideLinksByLinkId(linkId);
            sideLinkService.createSideLinks(link, recipientService.getRecipientsAndIfNotExistsCreateNew(linkDTO.getRecipients()));

            LinkDTO newDTO = linkMapper.linkToLinkDTO(linkRepository.save(link));
            newDTO.setRecipients(linkDTO.getRecipients());
            return newDTO;
        } else {
            throw new BadRequestException("Odkaz je sdílený, není možné ho upravit.");
        }
    }

    @Override
    public void deleteLink(Long linkId) {
        Link link = getLinkDetails(linkId);
        if (!link.isShared()) {
            log.debug("Delete Link by id: " + linkId + " from database.");
            link.getSideLinks()
                    .forEach(sideLink -> sideLink.getDownloadHistory()
                            .forEach(downloadHistory -> {
                                downloadHistory.setLink(null);
                                downloadHistory.setSideLink(null);
                            }));
            linkRepository.deleteById(linkId);
        } else {
            throw new BadRequestException("Odkaz je sdílený, není možné ho upravit.");
        }
    }


    // *********** CONDITIONS FOR ENDPOINTS
    @Override
    public boolean isDownloadsCountLimitExceeded(Link link) {
        if ((link.getDownloadsLimit() == 0) || (link.getDownloadsCount() < link.getDownloadsLimit())) {
            return false;
        } else {
            log.error("Download count limit exceeded on file id: " + link.getFile().getId());
            return true;
        }
    }

    @Override
    public boolean isValidUntilDateExceeded(Link link) {
        if (link.getValidUntil().isAfter(LocalDate.now())) {
            return true;
        } else {
            log.error("Download date until limit exceeded on file id: " + link.getFile().getId());
            return false;
        }
    }

    @Override
    public LinkValidationDTO isAnyLinkFileIncorrect(Link link) {
        DeleteDateSimpleDTO dto = deleteDateService.getDateWhenFileWillBeDeleted(link.getFile());

        if (dto.getNeverDelete() || dto.getDeleteAfterLimitExceeded()) {
            return new LinkValidationDTO(false, null);
        } else {
            return new LinkValidationDTO(link.getValidUntil().isAfter(dto.getDayOfDelete()), dto.getDayOfDelete());
        }
    }


    // *********** SCHEDULING
    @Override
    @Transactional
    public void disableDeprecatedLinks() {
        log.debug("Checking of deprecated links started.");
        getDeprecatedLinks().forEach(link -> {
            log.debug("Deprecated link on id: " + link.getId() + " will be disabled.");
            disableLink(link.getId());
        });
    }

    @Override
    public void setPlusActualDownloadsCount(Link link) {
        link.setDownloadsCount(link.getDownloadsCount() + 1);
        log.debug("Set Download count to " + link.getDownloadsCount());
        linkRepository.save(link);
    }


    // *********** PRIVATE - SAVE
    private List<Link> getDeprecatedLinks() {
        log.debug("Get all shared links, where their validUntil is before current time");

        return linkRepository.findAllBySharedIsTrue().stream()
                .filter(link -> link.getValidUntil().isBefore(LocalDate.now()))
                .collect(Collectors.toList());
    }

    private LinkDTO saveLinkAndReturnDTO(Link link) {
        return linkMapper.linkToLinkDTO(linkRepository.save(link));
    }

    private LinkDTO saveNewLinkAndReturnDTO(Link link, List<String> recipientMails) {
        link = linkRepository.save(link);
        fileService.addNewLinkToFileDM(link);

        List<Recipient> recipientList = recipientService.getRecipientsAndIfNotExistsCreateNew(recipientMails);

        sideLinkService.createSideLinks(link, recipientList);

        return linkMapper.linkToLinkDTO(link);
    }


}
