package cz.neit.dm.services;

import cz.neit.dm.entity.Link;
import cz.neit.dm.entity.Recipient;
import cz.neit.dm.entity.SideLink;
import cz.neit.dm.exceptions.ResourceNotFoundException;
import cz.neit.dm.io.LinkManager;
import cz.neit.dm.repositories.SideLinkRepository;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Service;

import java.util.List;
import java.util.stream.Collectors;

/**
 * Created by jhruby on 19.09.2018
 */
@Slf4j
@Service
@RequiredArgsConstructor
public class SideLinkServiceImpl implements SideLinkService {

    private final SideLinkRepository sideLinkRepository;
    private final LinkManager linkManager;
    private final RecipientService recipientService;


    // *********** GETTERS LIST
    @Override
    public List<SideLink> getSideLinksByLinkId(Long linkId) {
        log.debug("Getting side links by link id: " + linkId);
        return sideLinkRepository.findAllByLinkId(linkId);
    }


    // *********** GETTERS
    @Override
    public SideLink getSideLinkByToken(String sideLinkToken) {
        return sideLinkRepository.findByToken(sideLinkToken)
                .orElseThrow(() -> new ResourceNotFoundException("SideLink not found"));
    }

    @Override
    public SideLink getSideLinkByLinkIdAndRecipientId(Long linkId, Long recipientId) {
        log.debug("Getting side links by linkId: " + linkId + " and recipientId: " + recipientId);
        return sideLinkRepository.findByLinkIdAndRecipientId(linkId, recipientId)
                .orElseThrow(() -> new ResourceNotFoundException("SideLink not found"));
    }


    // *********** CRUD
    @Override
    public List<SideLink> createSideLinks(Link link, List<Recipient> recipientList) {
        return recipientList.stream()
                .map(recipient -> {
                    log.debug("Saving and creating side links for link on id: " + link.getId() + " with mail: " + recipient.getEmail());
                    SideLink sideLink = new SideLink();
                    sideLink.setLink(link);
                    sideLink.setRecipient(recipient);
                    recipientService.addNewSideLinkToRecipient(saveSideLink(sideLink));
                    return sideLink;
                })
                .collect(Collectors.toList());
    }

    @Override
    public void deleteSideLinksByLinkId(Long linkId) {
        sideLinkRepository.deleteAll(sideLinkRepository.findAllByLinkId(linkId));
    }


    // *********** SHARE OR DISABLE LINK FOR CUSTOMERS
    @Override
    public void shareSideLink(Long linkId) {
        log.debug("Set share for sideLinks on linkId: " + linkId );

        getSideLinksByLinkId(linkId)
                .forEach(sideLink -> {
                    sideLink.setToken(linkManager.getNewUuid4());
                    saveSideLink(sideLink);
                });
    }

    @Override
    public void disableSideLink(Long linkId) {
        log.debug("Disable sharing for sideLinks on linkId: " +linkId);

        getSideLinksByLinkId(linkId)
                .forEach(sideLink -> {
                    sideLink.setToken(null);
                    saveSideLink(sideLink);
                });
    }


    // *********** PRIVATE - SAVE
    private SideLink saveSideLink(SideLink sideLink) {
        return sideLinkRepository.save(sideLink);
    }


}
