package cz.neit.dm.services;

import cz.neit.dm.dtos.mappers.GroupMapper;
import cz.neit.dm.dtos.mappers.UserMapper;
import cz.neit.dm.dtos.modeldto.GroupSimpleDTO;
import cz.neit.dm.dtos.modeldto.UserDTO;
import cz.neit.dm.dtos.modeldto.UserGroupsDTO;
import cz.neit.dm.dtos.modeldto.UserSimpleDTO;
import cz.neit.dm.entity.PasswordResetToken;
import cz.neit.dm.entity.User;
import cz.neit.dm.entity.enums.Language;
import cz.neit.dm.exceptions.BadRequestException;
import cz.neit.dm.exceptions.ResourceNotFoundException;
import cz.neit.dm.repositories.GroupMembershipRepository;
import cz.neit.dm.repositories.PasswordTokenRepository;
import cz.neit.dm.repositories.UserRepository;
import cz.neit.dm.security.acl.Roles;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.security.crypto.bcrypt.BCryptPasswordEncoder;
import org.springframework.stereotype.Service;

import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

import static cz.neit.dm.entity.enums.MembershipType.MANAGER;

/**
 * Created by jhruby on 25.07.2018
 */
@Slf4j
@Service
@RequiredArgsConstructor
public class UserServiceImpl implements UserService {

    private final UserRepository userRepository;
    private final GroupMembershipRepository groupMembershipRepository;
    private final GroupMapper groupMapper;
    private final UserMapper userMapper;
    private final BCryptPasswordEncoder bCryptPasswordEncoder;
    private final PasswordTokenRepository passwordTokenRepository;
    private final FileService fileService;

    @Value("${app.default.language}")
    private String BASE_DEFAULT_LANGUAGE;

    @Value("${app.mail.expire.token.hour}")
    private int expireToken;

    // *********** GETTERS LIST
    @Override
    public List<UserDTO> getAllUsers() {
        log.debug("Getting All Users");

        List<User> userList = new ArrayList<>();
        userRepository.findAll().forEach(userList::add);

        return userList.stream()
                .map(user -> {
                    UserDTO userDTO = userMapper.userToUserDTO(user);
                    userDTO.setGroups(setUserGroupDTO(user.getId()));
                    return userDTO;
                })
                .collect(Collectors.toList());
    }

    @Override
    public List<UserSimpleDTO> getAllSimpleUsers() {
        log.debug("Getting All Users");

        List<User> userList = new ArrayList<>();
        userRepository.findAll().forEach(userList::add);

        return userList.stream()
                .map(userMapper::userToSimpleUserDTO)
                .collect(Collectors.toList());
    }

    // *********** USER SEARCHING
    @Override
    public List<UserDTO> getUsersByFullNameOnGroupId(String fullname, Long groupId) {
        //returning users that aren't in group yet
        List<User> userList = userRepository.findAllByFullnameLikeIgnoreCase("%"+fullname+"%");

        List<User> usersInGroup = new ArrayList<>();
        groupMembershipRepository.findAllByGroupId(groupId)
                .forEach(groupMembership -> usersInGroup.add(groupMembership.getUser()));

        return userList.stream()
                .filter(user -> !usersInGroup.contains(user))
                .map(userMapper::userToUserDTO)
                .collect(Collectors.toList());
    }


    // *********** GETTERS
    @Override
    public User getUserDetail(Long userId) {
        return userRepository.findById(userId)
                .orElseThrow(() -> new ResourceNotFoundException("Uživatel nebyl nalezen."));
    }

    @Override
    public UserDTO getUserDetailDTO(Long id) {
        log.debug("Getting User by id: " + id);

        Optional<UserDTO> userDTO = userRepository.findById(id)
                .map(user -> {
                    user.getGroupMemberships().size();
                    UserDTO u = userMapper.userToUserDTO(user);
                    u.setGroups(setUserGroupDTO(user.getId()));
                    return u;
                });
        if (!userDTO.isPresent()){
            throw new ResourceNotFoundException("Uživatel nebyl nalezen.");
        }
        return userDTO.get();
    }

    @Override
    public User getUserByEmail(String email) {
        return userRepository.findByEmail(email)
                .orElseThrow(() -> new BadRequestException("Uživatel  emailem '" + email +"' nebyl nalezen."));
    }


    // *********** CRUD
    @Override
    public UserDTO createUser(UserDTO userDTO) {
        log.debug("Creating new User, name: " + userDTO.getFullname());
        User user = userMapper.userDtoToUser(userDTO);
        user.setLanguage(Language.valueOf(BASE_DEFAULT_LANGUAGE));
        user.setRole(Roles.USER);
        user.setPassword(bCryptPasswordEncoder.encode(userDTO.getPassword()));
        user.setFullname(user.getFirstname(),user.getLastname());

        if (userRepository.findByUsername(user.getUsername()).isPresent()) {
            throw new BadRequestException("Uživatelské jméno již existuje");
        }

        if (userRepository.findByEmail(user.getEmail()).isPresent()) {
            throw new BadRequestException("Emailová adresa je již použita");
        }

        userRepository.save(user);
        return userMapper.userToUserDTO(user);
    }

    @Override
    public PasswordResetToken createPasswordResetToken(User user, String token) {
        PasswordResetToken myToken = new PasswordResetToken();
        myToken.setUser(user);
        myToken.setToken(token);
        myToken.setExpiryDate(LocalDateTime.now().plusHours(expireToken));
        return passwordTokenRepository.save(myToken);
    }

    @Override
    public void deletePasswordResetToken(String token) {
        PasswordResetToken myToken = passwordTokenRepository.findByToken(token).get();

        if (passwordTokenRepository.findByToken(token).isPresent()) {
            passwordTokenRepository.delete(myToken);
        }


    }

    @Override
    public UserSimpleDTO updateUser(Long id, UserSimpleDTO userSimpleDTO) {
        log.debug("Updating User, id: " + id);

        User user = getUserDetail(id);
        user.setEmail(userSimpleDTO.getEmail());
        user.setFirstname(userSimpleDTO.getFirstname());
        user.setLastname(userSimpleDTO.getLastname());
        user.setRole(userSimpleDTO.getRole());

        return saveUserAndReturnDTO(user);
    }

    @Override
    public void deleteUser(Long id) {
        log.debug("Deletting User, id: "+ id);
        if(!fileService.getFilesByAuthorId(id).isEmpty()) {
            throw new BadRequestException("Uživatel je autorem soubrů, předejte jejich vlastnictví.");
        }
        userRepository.deleteById(id);
    }


    @Override
    public UserSimpleDTO changeAuthorOnFilesWithoutGroup(Long userId, Long newAuthorId) {
        User newAuthor = getUserDetail(newAuthorId);

        fileService.getFilesByUserIdWithoutGroup(userId)
                .forEach(fileDM -> {
                    fileDM.getLinks().forEach(link -> {
                        if (link.getAuthor().getUsername().equals(fileDM.getAuthor().getUsername())) {
                            link.setAuthor(newAuthor);
                        }
                    });
                    fileDM.setAuthor(newAuthor);
                    fileService.save(fileDM);
                });

        return userMapper.userToSimpleUserDTO(newAuthor);
    }

    // *********** VALIDATION
    @Override
    public PasswordResetToken validatePasswordResetToken(String token) {
        Optional<PasswordResetToken> passwordResetToken = passwordTokenRepository.findByToken(token);

        if (!passwordResetToken.isPresent()) {
            log.warn("Token "+ token + " is invalid.");
            throw new BadRequestException("Máte nevalidní token. Opakujte změnu hesla.");
        }

        if (passwordResetToken.get().getExpiryDate().isBefore(LocalDateTime.now())) {
            log.warn("Token " + token +" was expired!");
            throw new BadRequestException("Vypršela platnost odkazu. Opakujte změnu hesla.");
        }
        return passwordResetToken.get();
    }


    @Override
    public UserSimpleDTO saveUserAndReturnDTO(User user) {
        user.setFullname(user.getFirstname(),user.getLastname());
        userRepository.save(user);
        return userMapper.userToSimpleUserDTO(user);
    }

    @Override
    public void simpleSaveUser(User user) {
        userRepository.save(user);
    }


    // *********** PRIVATE - SUPPORTIVE METHODS **********
    private UserGroupsDTO setUserGroupDTO(Long userId) {
        UserGroupsDTO userGroupsDTO = new UserGroupsDTO();
        List<GroupSimpleDTO> userOfList = new ArrayList<>();
        List<GroupSimpleDTO> managerOfList = new ArrayList<>();

        groupMembershipRepository.findAllByUserId(userId).forEach(
                pair -> {
                    if (pair.getMembershipType().equals(MANAGER)) {
                        managerOfList.add(groupMapper.groupToSimpleGroupDTO(pair.getGroup()));
                    } else {
                        userOfList.add(groupMapper.groupToSimpleGroupDTO(pair.getGroup()));
                    }
                });

        userGroupsDTO.setUserOf(userOfList);
        userGroupsDTO.setManagerOf(managerOfList);

        return userGroupsDTO;
    }
}
